import {
    FrostBaseEntity,
    FrostDatastreamInput,
    FrostFeatureOfInterest,
    FrostLocation,
    FrostObservationInput,
    FrostProjection,
    FrostThing
} from "./DataTypes";
import fetch from "node-fetch";
import {FlowResponse} from "../here/DataTypes";
import {info} from "../logger/logger";

const THINGS_PATH = "Things";
const LOCATIONS_PATH = "Locations";
const FEATURES_OF_INTEREST_PATH = "FeaturesOfInterest";
const DATASTREAMS_PATH = "Datastreams";
const OBSERVATIONS_PATH = "Observations";
const SENSORS_PATH = "Sensors";
const OBSERVED_PROPERTIES_PATH = "ObservedProperties";

type ItemType = "location" | "thing" | "featureOfInterest" | "datastream" | "observation" | "sensor" | "observedProperty";

const getPathFromType = (type: ItemType) => {
    switch (type) {
        case "thing":
            return THINGS_PATH;
        case "location":
            return LOCATIONS_PATH;
        case "featureOfInterest":
            return FEATURES_OF_INTEREST_PATH;
        case "datastream":
            return DATASTREAMS_PATH;
        case "observation":
            return OBSERVATIONS_PATH;
        case "sensor":
            return SENSORS_PATH;
        case "observedProperty":
            return OBSERVED_PROPERTIES_PATH;
    }

    throw new Error(`Unknown type ${type}!`);
};

const DEFAULT_PROJECTION: FrostProjection = ["@iot.id"];

export class FrostClient {

    constructor(
        private baseUrl: string
    ) {
        // Initialize Here sensor and Speed observed property if they are not already inserted
        this.insertItemIfNotExists("sensor", {
            "@iot.id": "here",
            name: "Here",
            description: "Here - Traffic Flow API",
            encodingType: "text/html",
            metadata: "https://developer.here.com/documentation/traffic/dev_guide/topics_v6.1/example-flow.html"
        })
            .then(inserted => {
                if (inserted) console.log("Here sensor was inserted.");
            })
            .catch(err => console.error(err));
        this.insertItemIfNotExists("observedProperty", {
            "@iot.id": "speed",
            name: "Speed",
            definition: "http://www.qudt.org/qudt/owl/1.0.0/quantity/Instances.html#Speed",
            description: "The Speed."
        })
            .then(inserted => {
                if (inserted) console.log("Speed observed property was inserted.");
            })
            .catch(err => console.error(err));
    }

    /**
     * Returns the first thing with the specified id, or null if not found.
     * @param id
     * @param projection
     */
    async getSingleThingById(id: string, projection?: FrostProjection): Promise<FrostThing> {
        return this.getSingleItemById("thing", id, projection);
    }

    /**
     * Returns the first location with the specified id, or null if not found.
     * @param id
     * @param projection
     */
    async getSingleLocationById(id: string, projection?: FrostProjection): Promise<FrostLocation> {
        return this.getSingleItemById("location", id, projection);
    }

    /**
     * Returns the first feature of interest with the specified id, or null if not found.
     * @param id
     * @param projection
     */
    async getSingleFeatureOfInterestById(id: string, projection?: FrostProjection): Promise<FrostFeatureOfInterest> {
        return this.getSingleItemById("featureOfInterest", id, projection);
    }

    async getSingleDatastreamById(id: string, projection?: FrostProjection): Promise<any> {
        return this.getSingleItemById("datastream", id, projection);
    }

    /**
     * Returns the number of inserted observations
     * @param hereFlow
     * @param forceUpdateLocations
     */
    async mapAndInsertHereFlowResponse(hereFlow: FlowResponse, forceUpdateLocations: boolean = false): Promise<number> {
        const data = hereFlow.RWS
            .flatMap(rws => rws.RW)
            .flatMap(rw => {
                const timeStamp = rw.PBT;
                const locationsAndResults =
                    rw.FIS
                        .flatMap(fis => fis.FI)
                        .map(fi => {
                            const thingId = `${rw.LI}-${fi.TMC.DE}-${fi.TMC.PC}`;
                            const thingDescription = `${rw.DE} direction to ${fi.TMC.DE}`;
                            const thing: FrostThing = {
                                "@iot.id": thingId,
                                name: thingId,
                                description: thingDescription
                            };
                            const location: FrostLocation = {
                                "@iot.id": thingId,
                                name: thingId,
                                description: thingDescription,
                                encodingType: "application/vnd.geo+json",
                                location: {
                                    type: "MultiLineString",
                                    coordinates: fi.SHP!
                                        .map(v => v.value.toString())
                                        .map(
                                            str =>
                                                str
                                                    .trim()
                                                    .split(" ")
                                                    .map(coord => coord.split(",").reverse().map(x => +x))
                                        )
                                },
                                properties: {
                                    length: fi.TMC.LE //In Kilometers
                                },
                                Things: [{
                                    "@iot.id": thingId
                                }],
                            };
                            const featureOfInterest: FrostFeatureOfInterest = {
                                "@iot.id": location["@iot.id"],
                                name: location.name,
                                description: location.description,
                                encodingType: location.encodingType,
                                feature: location.location,
                                properties: location.properties,
                            };
                            const speed = fi.CF[0]?.SP;
                            const confidence = fi.CF[0]?.CN;
                            const jamFactor = fi.CF[0]?.JF;
                            return {thing, location, featureOfInterest, timeStamp, speed, confidence, jamFactor};
                        })
                        .filter(data => !!data.speed); // if failed to obtain speed data, don't consider it
                /***** Log if ID repetitions Found *****/
                Object.entries(locationsAndResults.map(data => data.featureOfInterest["@iot.id"]!)
                    .reduce(function (prev: { [key: string]: number }, cur: string) {
                        prev[cur] = (prev[cur] || 0) + 1;
                        return prev;
                    }, {}))
                    .forEach(([foiId, repetitions]) => {
                        if (repetitions > 1) {
                            console.error(`FeaturesOfInterest: ${foiId} is repeated ${repetitions} times`);
                        }
                    });
                /***** *************************** *****/
                return locationsAndResults;
            });
        return Promise.all(data.map(roadwayData => {
            const datastreamId = `here-${roadwayData.thing["@iot.id"]}`;
            return this.insertItemIfNotExists("thing", roadwayData.thing)
                .then(() => {
                    const datastream: FrostDatastreamInput = {
                        "@iot.id": datastreamId,
                        name: datastreamId,
                        description: `Speed readings from Here for roadway ${roadwayData.thing.description}`,
                        observationType: "http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_Measurement",
                        unitOfMeasurement: {
                            name: "Kilometers per hour",
                            symbol: "km/h",
                            definition: "http://www.qudt.org/qudt/owl/1.0.0/unit/Instances.html#KilometerPerHour"
                        },
                        ObservedProperty: {
                            "@iot.id": "speed"
                        },
                        Sensor: {
                            "@iot.id": "here"
                        },
                        Thing: {
                            "@iot.id": roadwayData.thing["@iot.id"]
                        },
                    };
                    return this.insertItemIfNotExists("datastream", datastream);
                })
                .then(() => this.insertItemIfNotExists("location", roadwayData.location, forceUpdateLocations))
                .then(() => this.insertItemIfNotExists("featureOfInterest", roadwayData.featureOfInterest, forceUpdateLocations))
                .then(() => {
                    const observation: FrostObservationInput = {
                        phenomenonTime: roadwayData.timeStamp,
                        result: roadwayData.speed,
                        resultQuality: roadwayData.confidence,
                        resultTime: roadwayData.timeStamp,
                        FeatureOfInterest: {
                            "@iot.id": roadwayData.featureOfInterest["@iot.id"]
                        },
                        Datastream: {
                            "@iot.id": datastreamId
                        },
                        parameters: {
                            jamFactor: roadwayData.jamFactor
                        }
                    };
                    return this.insertItem("observation", observation);
                })
                .then(() => 1);
        })).then(length => length.reduce((a, b) => a + b, 0));
    }

    private async insertItemIfNotExists(
        type: ItemType,
        frostItemWithId: FrostBaseEntity,
        forceUpdate: boolean = false
    ): Promise<boolean> {
        const existingThing = await this.getSingleItemById(type, frostItemWithId["@iot.id"]!);
        if (existingThing && forceUpdate) {
            info(`Force updating item ${frostItemWithId["@iot.id"]} with type ${type}`);
            return this.updateItem(type, frostItemWithId);
        }
        if (existingThing)
            return false;
        return this.insertItem(type, frostItemWithId);
    }

    private async insertItem(
        type: ItemType,
        frostItem: FrostBaseEntity
    ): Promise<boolean> {
        // if (type == "location")
        //     console.debug(`inserting item with type ${type}`);
        const path = getPathFromType(type);
        let url = `${this.baseUrl}/${path}`;
        return fetch(url, {
            method: "POST",
            body: JSON.stringify(frostItem)
        })
            .then(res => {
                if (!res.ok) {
                    console.error(`Error inserting item ${JSON.stringify(frostItem)} with type ${type}`);
                    res.text().then(text => console.error(text));
                    const err = new Error(`Error inserting item ${JSON.stringify(frostItem)} with type ${type}`);
                    err.name = "COULD_NOT_INSERT_ITEM";
                    throw err;
                }
                return res.ok;
            });
    }

    private async updateItem(
        type: ItemType,
        frostItem: FrostBaseEntity
    ): Promise<boolean> {
        // if (type == "location")
        //     console.debug(`inserting item with type ${type}`);
        const path = getPathFromType(type);
        const encodedId = frostItem["@iot.id"]?.split("/").map(x => encodeURIComponent(x)).join("/");
        let url = `${this.baseUrl}/${path}('${encodedId}')`;
        return fetch(url, {
            method: "PATCH",
            body: JSON.stringify(frostItem)
        })
            .then(res => {
                if (!res.ok) {
                    console.error(`Error updating item ${JSON.stringify(frostItem)} with type ${type}`);
                    res.text().then(text => console.error(text));
                    const err = new Error(`Error inserting item ${JSON.stringify(frostItem)} with type ${type}`);
                    err.name = "COULD_NOT_INSERT_ITEM";
                    throw err;
                }
                return res.ok;
            });
    }

    private async getSingleItemById(
        type: ItemType,
        id: string,
        projection: FrostProjection = DEFAULT_PROJECTION
    ): Promise<FrostBaseEntity> {
        const path = getPathFromType(type);
        const params = {
            "$filter": `id eq '${id}'`,
            "$top": 1,
            ...(projection ? {"$select": projection?.join(",")} : {})
        };
        return this.get(path, params).then(res => res.value[0]);
    }

    private async get(urlPath: string, params?: { [key: string]: string | number }): Promise<any> {
        let url = `${this.baseUrl}/${urlPath}`;
        if (!params)
            params = {};
        if (Object.keys(params)) {
            url += "?";
            url += Object.entries(params)
                .map(([key, value]) => [key, encodeURIComponent(value)])
                .map(([key, value]) => `${key}=${value}`)
                .join("&");
        }
        return fetch(url)
            .then(res => res.json())
            .then(json => {
                // console.debug(`Getting url "${url}"`);
                // console.debug(json);
                return json;
            });
    }
}
